@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2006 - 2010 René Rebe
+ * Copyright (C) 2006 - 2016 René Rebe, ExactCODE GmbH
  *           (C) 2006, 2007 Archivista GmbH, CH-8042 Zuerich
  * 
  * This program is free software; you can redistribute it and/or modify
@@ -21,6 +21,7 @@
 #include <string.h>
 
 #include <string>
+#include <vector>
 #include <iostream>
 #include <fstream>
 
@@ -142,7 +143,7 @@ boolean fill_input_buffer (j_decompress_ptr cinfo)
   src->pub.bytes_in_buffer = nbytes;
   src->start_of_file = FALSE;
   
-  return TRUE;
+  return (boolean)TRUE;
 }
 
 
@@ -178,7 +179,7 @@ void cpp_stream_src (j_decompress_ptr cinfo, std::istream* stream)
   if (cinfo->src == NULL) {	/* first time for this JPEG object? */
     cinfo->src = (jpeg_source_mgr*) malloc (sizeof(cpp_src_mgr));
     src = (cpp_src_mgr*) cinfo->src;
-    src->buffer = (JOCTET *) malloc (INPUT_BUF_SIZE * sizeof(JOCTET));
+    src->buffer = (JOCTET*) malloc (INPUT_BUF_SIZE * sizeof(JOCTET));
   }
 
   src = (cpp_src_mgr*) cinfo->src;
@@ -212,7 +213,7 @@ void init_destination (j_compress_ptr cinfo)
   cpp_dest_mgr* dest = (cpp_dest_mgr*) cinfo->dest;
 
   /* Allocate the output buffer --- it will be released when done with image */
-  dest->buffer = (JOCTET *)
+  dest->buffer = (JOCTET*)
       (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
 				  OUTPUT_BUF_SIZE * sizeof(JOCTET));
 
@@ -231,7 +232,7 @@ boolean empty_output_buffer (j_compress_ptr cinfo)
   dest->pub.next_output_byte = dest->buffer;
   dest->pub.free_in_buffer = OUTPUT_BUF_SIZE;
 
-  return TRUE;
+  return (boolean)TRUE;
 }
 
 
@@ -275,42 +276,140 @@ void cpp_stream_dest (j_compress_ptr cinfo, std::ostream* stream)
 /* *** back on-topic *** */
 
 JPEGCodec::JPEGCodec (Image* _image)
-  : ImageCodec (_image)
+  : ImageCodec (_image), colorspace(JCS_UNKNOWN)
 {
 }
 
-int JPEGCodec::readImage (std::istream* stream, Image& image, const std::string& decompres)
+int JPEGCodec::readImage (std::istream* stream, Image& image, const std::string& decompress)
 {
-  if (stream->peek () != 0xFF)
+  Args args(decompress);
+
+  if (stream->peek() != 0xFF)
     return false;
-  stream->get (); // consume silently
-  if (stream->peek () != 0xD8)
+  stream->get(); // consume silently
+  if (stream->peek() != 0xD8)
     return false;
   
   if (0) { // TODO: differentiate JFIF vs. Exif?
     // quick magic check
     char buf [10];
-    stream->read (buf, sizeof (buf));
-    stream->seekg (0);
+    stream->read(buf, sizeof(buf));
+    stream->seekg(0);
     
     if (buf[6] != 'J' || buf[7] != 'F' || buf[8] != 'I' || buf[9] != 'F')
       return false;
   }
   
-  if (!readMeta (stream, image))
-    return false;
+  uint16_t height = 0;
+  {
+    std::string arg = args.containsPrefixedAndRemove("height=");
+    if (!arg.empty()) {
+      std::stringstream s(arg);
+      // TODO: parsing error handling, sigh!
+      s >> height;
+    }
+  }
   
-  // on-demand compression
-  image.setRawData (0);
+  JPEGCodec* codec = 0;
+  image.setRawData(0); // on-demand compression
   
-  // freestanding instance
-  JPEGCodec* codec = new JPEGCodec(&image);
-  image.setCodec(codec);
+  if (height == 0) {
+    if (!readMeta(stream, image)) {
+      return false;
+    }
+    codec = new JPEGCodec(&image); // freestanding instance
+    image.setCodec(codec);
+    
+    stream->clear(); // private copy for deferred decoding
+    stream->seekg(0);
+    *stream >> codec->private_copy.rdbuf();
+  } else {
+    codec = new JPEGCodec(&image); // freestanding instance
+    
+    // scan thru segments and potentially update height, sigh!
+    {
+      std::vector<uint8_t> buffer;
+      stream->seekg(0);
+      bool found = false;
+      while (stream->good() && !found) {
+	buffer.resize(2);
+	stream->read((char*)&buffer[0], 2);
+	if (buffer[0] != 0xff) {
+	  std::cerr << "not a tag" << std:: endl;
+	  stream->seekg(0);
+	  return false;
+	}
+	
+	switch (buffer[1]) {
+	  // types wo/ data
+	case 0xd9: // EOI
+	  found = true; // cheating
+	  std::cerr << "EOI wo/ SOF segment?" << std::endl;
+	  
+	case 0xd8: // SOI
+	case 0xd0: case 0xd1: case 0xd2: case 0xd3: // RSTn
+	case 0xd4: case 0xd5: case 0xd6: case 0xd7:
+	  break;
+	
+	  // type w/ variable length
+	case 0xc0: // SOF0
+	case 0xc2: // SOF2
+	  found = true;
+	case 0xe0: case 0xe1:	case 0xe2: case 0xe3: // APPn
+	case 0xe4: case 0xe5:	case 0xe6: case 0xe7:
+	case 0xe8: case 0xe9:	case 0xea: case 0xeb:
+	case 0xec: case 0xed:	case 0xee: case 0xef:
+
+	
+	case 0xc4: // DHT
+	case 0xdb: // DQt
+	case 0xda: // SOS
+	case 0xfe: // COM
+	case 0xdd: // DRI
+	  {
+	    buffer.resize(4, 0);
+	    stream->read((char*)&(buffer[2]), 2);
+	    uint16_t len = buffer[2] << 8 | buffer[3];
+	    //std::cerr << "len: " << len << std::endl,
+	    buffer.resize(2 + len);
+	    stream->read((char*)&(buffer[4]), len - 2);
+	    
+	    if (found) {
+	      len = buffer[5] << 8 | buffer[6];
+	      if (len == 0xffff || len != height) {
+		std::cerr << "Updating JPEG height to: " << height << " (was: " << len << ")" << std::endl;
+		buffer[5] = height >> 8;
+		buffer[6] = height & 0xff;
+	      }
+	    }
+	  }
+	  break;
+	  
+	default:
+	  std::cerr << "Unsupported segment type: " << std::hex << (unsigned)buffer[1] << std::dec
+		    << " not setting Height!" << std:: endl;
+	  found = true; // cheating to end loop
+	  break; // try decoding without altered height
+	}
+	codec->private_copy.write((char*)&buffer[0], buffer.size());
+      }
+      
+      // copy the rest
+      *stream >> codec->private_copy.rdbuf();
+    }
+
+    if (!readMeta(&codec->private_copy, image)) {
+      delete codec;
+      return false;
+    }
+    
+    image.setCodec(codec);
+  }
   
-  // private copy for deferred decoding
-  //private_copy.str().resize(stream->tellg());
-  stream->seekg (0);
-  *stream >> codec->private_copy.rdbuf();
+  if (args.containsAndRemove("ycck"))
+    codec->colorspace = JCS_YCCK;
+  else if (args.containsAndRemove("rgb"))
+    codec->colorspace = JCS_RGB;
   
   // parse Exif data, might contain non-identifiy orientation transform
   codec->parseExif(image);
@@ -321,34 +420,44 @@ int JPEGCodec::readImage (std::istream* stream, Image& image, const std::string&
 bool JPEGCodec::writeImage (std::ostream* stream, Image& image, int quality,
 			    const std::string& compress)
 {
-  std::string c (compress);
-  std::transform (c.begin(), c.end(), c.begin(), tolower);
-
+  Args args(compress);
+  const bool debug = args.containsAndRemove("debug");
+  
   // if the instance is freestanding it can only be called by the mux
   // if the cache is valid
-  if (_image && c != "recompress") {
+  if (_image && !args.containsAndRemove("recompress")) {
     // if meta information was modified re-encode the stream
     if (image.isMetaModified()) {
-      std::cerr << "Re-encoding DCT coefficients (due meta changes)." << std::endl;
+      if (debug)
+	std::cerr << "Re-encoding DCT coefficients (due meta changes)." << std::endl;
       doTransform (JXFORM_NONE, image, stream);
-    } else {
-      std::cerr << "Writing unmodified DCT buffer." << std::endl;
+    } else if (stream) {
+      if (debug)
+	std::cerr << "Writing unmodified DCT buffer." << std::endl;
       *stream << private_copy.str();
     }
     
     return true;
   }
   
+  if (image.w <= 0 || image.h <= 0) {
+    std::cerr << "Can not write image with 0 dimension" << std::endl;
+    return false;
+  }
+  
+  JPEGCodec* cache =
+    args.containsAndRemove("cache") ? new JPEGCodec(&image) : 0;
+  if (cache)
+    image.setCodec(cache);
+  
   // really encode
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
 
-  JSAMPROW buffer[1];	/* pointer to JSAMPLE row[s] */
-
-  /* Initialize the JPEG compression object with default error handling. */
+  // Initialize the JPEG compression object with default error handling.
   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_compress(&cinfo);
-  cpp_stream_dest (&cinfo, stream);
+  cpp_stream_dest(&cinfo, cache ? &cache->private_copy : stream);
   
   cinfo.in_color_space = JCS_UNKNOWN;
   if (image.bps == 8 && image.spp == 3)
@@ -357,7 +466,7 @@ bool JPEGCodec::writeImage (std::ostream* stream, Image& image, int quality,
     cinfo.in_color_space = JCS_GRAYSCALE;
   else if (image.bps == 8 && image.spp == 4)
     cinfo.in_color_space = JCS_CMYK;
-
+  
   if (cinfo.in_color_space == JCS_UNKNOWN) {
     if (image.bps < 8)
       std::cerr << "JPEGCodec: JPEG can not hold less than 8 bit-per-channel." << std::endl;
@@ -372,30 +481,67 @@ bool JPEGCodec::writeImage (std::ostream* stream, Image& image, int quality,
   cinfo.input_components = image.spp;
   cinfo.data_precision = image.bps; 
   
-  /* defaults depending on in_color_space */
+  // defaults depending on in_color_space
   jpeg_set_defaults(&cinfo);
-  
   jpeg_compress_set_density (&cinfo, image);
-
-  jpeg_set_quality(&cinfo, quality, FALSE); /* do not limit to baseline-JPEG values */
-
-  /* Start compressor */
-  jpeg_start_compress(&cinfo, TRUE);
-
-  /* Process data */
-  while (cinfo.next_scanline < cinfo.image_height) {
-    buffer[0] = (JSAMPLE*) image.getRawData() +
-      cinfo.next_scanline*image.stride();
-    (void) jpeg_write_scanlines(&cinfo, buffer, 1);
+  jpeg_set_quality(&cinfo, quality, (boolean)FALSE); // do not limit to baseline-JPEG values
+  
+  // sub-sampling
+  if (cinfo.in_color_space == JCS_RGB) {
+    if (args.containsAndRemove("4:4:4")) {
+      cinfo.comp_info[0].h_samp_factor =
+	cinfo.comp_info[0].v_samp_factor =
+	cinfo.comp_info[1].h_samp_factor =
+	cinfo.comp_info[1].v_samp_factor =
+	cinfo.comp_info[2].h_samp_factor =
+	cinfo.comp_info[2].v_samp_factor = 1;
+    } else if (args.containsAndRemove("4:2:2")) {
+      cinfo.comp_info[0].h_samp_factor = 2;
+      cinfo.comp_info[0].v_samp_factor =
+	cinfo.comp_info[1].h_samp_factor =
+	cinfo.comp_info[1].v_samp_factor =
+	cinfo.comp_info[2].h_samp_factor =
+	cinfo.comp_info[2].v_samp_factor = 1;
+    } else if (args.containsAndRemove("4:1:1")) {
+      cinfo.comp_info[0].h_samp_factor =
+	cinfo.comp_info[0].v_samp_factor = 2;
+      cinfo.comp_info[1].h_samp_factor =
+	cinfo.comp_info[1].v_samp_factor =
+	cinfo.comp_info[2].h_samp_factor =
+	cinfo.comp_info[2].v_samp_factor = 1;
+    } 
   }
 
-  /* Finish compression and release memory */
+  if (!args.str().empty())
+    std::cerr << "JPEGCodec: Unrecognized encoding options '" << args.str() << "'" << std::endl;
+  
+  // Start compressor
+  jpeg_start_compress(&cinfo, (boolean)TRUE);
+
+  // Process data
+  while (cinfo.next_scanline < cinfo.image_height) {
+    JSAMPROW buffer[1]; // pointer to JSAMPLE row[s]
+    buffer[0] = (JSAMPLE*)image.getRawData() + cinfo.next_scanline * image.stride();
+    if (jpeg_write_scanlines(&cinfo, buffer, 1) < 1) {
+      std::cerr << "Could not write scanline." << std::endl;
+      jpeg_finish_compress(&cinfo);
+      jpeg_destroy_compress(&cinfo);
+      return false;
+    }
+  }
+
+  // Finish compression and release memory
   jpeg_finish_compress(&cinfo);
   jpeg_destroy_compress(&cinfo);
 
   if (jerr.num_warnings)
     std::cerr << jerr.num_warnings << " Warnings." << std::endl;
 
+  // if we cached a copy, write it to the actual stream, too
+  if (cache && stream) {
+    *stream << cache->private_copy.str();
+  }
+  
   return true;
 }
 
@@ -443,7 +589,11 @@ void JPEGCodec::parseExif (Image& image)
   }
 
   // Get the marker parameter length count
-  unsigned length = readExif<uint16_t>(exif_data + 2, true); // always big-endian
+  uint16_t length = readExif<uint16_t>(exif_data + 2, true); // always big-endian
+  if (length > exif_data_p.size()) {
+    std::cerr << "Exif header length limitted" << std::endl;
+    length = exif_data_p.size();
+  }
   
   // length includes itself, so must be at least 2 + Exif data length must be at least 6
   if (length < 8)
@@ -465,11 +615,9 @@ void JPEGCodec::parseExif (Image& image)
   
   // Check tag mark
   if (big_endian) {
-    if (exif_data[2] != 0) return;
-    if (exif_data[3] != 0x2A) return;
+    if (exif_data[2] != 0 || exif_data[3] != 0x2A) return;
   } else {
-    if (exif_data[3] != 0) return;
-    if (exif_data[2] != 0x2A) return;
+    if (exif_data[3] != 0 || exif_data[2] != 0x2A) return;
   }
 
   // get first IFD offset (offset to IFD0)
@@ -482,18 +630,84 @@ void JPEGCodec::parseExif (Image& image)
   offset += 2;
 
   // search for orientation tag in IFD0
-  for (;;) {
-    if (offset > length - 12) return; // check end of data segment
+  uint16_t orientation = 0, unit = 0;
+  uint32_t xres = 0, yres = 0;
+  
+  for (; number_of_tags > 0; --number_of_tags, offset += 12) {
+    if (offset > length - 12) break; // check end of data segment
+    
     // get tag number
-    unsigned tagnum = readExif<uint16_t>(exif_data + offset, big_endian);
-    if (tagnum == 0x0112) break; // orientation tag
-    if (--number_of_tags == 0) return;
-    offset += 12;
+    uint16_t tag = readExif<uint16_t>(exif_data + offset, big_endian);
+    uint16_t type = readExif<uint16_t>(exif_data + offset + 2, big_endian);
+    uint32_t count = readExif<uint32_t>(exif_data + offset + 4, big_endian);
+    uint32_t value = readExif<uint32_t>(exif_data + offset + 8, big_endian);
+    
+    //std::cerr << std::hex << tag << std::dec << " " << type << " " << count << " " << value << std::endl;
+    
+    // global range check
+    if ((type == 5 || type == 10) && (value + 4 >= length) || // RATIONAL
+        (type == 2 && count > 4 && value + count >= length)) // ASCII, could be short
+    {
+	std::cerr << "Exif tag index out of range, skipped." << std::endl;
+	continue;
+    }
+    
+    if (tag == 0x011a) // xres
+    {
+      uint32_t x = readExif<uint32_t>(exif_data + value, big_endian),
+	y = readExif<uint32_t>(exif_data + value + 4, big_endian);
+      xres = (double)x/y;
+    } else if (tag == 0x011b) // yres
+    {
+      uint32_t x = readExif<uint32_t>(exif_data + value, big_endian),
+        y = readExif<uint32_t>(exif_data + value + 4, big_endian);
+      yres = (double)x/y;
+    } else if (tag == 0x0128) // unit
+    {
+      uint16_t u = readExif<uint16_t>(exif_data + offset + 8, big_endian);
+      if (unit != 0)
+	std::cerr << "Exif unit already set?" << std::endl;
+
+      if (u == 2 || u == 3) // inch, cm
+	unit = u;
+      else
+	std::cerr << "Exif unit invalid: " << u << std::endl;
+    }
+    
+    if (tag == 0x0112) // orientation tag
+    {
+      uint16_t o = readExif<uint16_t>(exif_data + offset + 8, big_endian);
+      if (orientation != 0)
+        std::cerr << "Exif orientation already set?" << std::endl;
+     
+      if (o <= 8)
+	orientation = o;
+      else
+	std::cerr << "Exif orientation invalid: " << o << std::endl;
+    }
   }
 
-  // get the orientation value
-  unsigned orientation = readExif<uint16_t>(exif_data + offset + 8, big_endian);
-  if (orientation > 8) return;
+  if (xres || yres)
+  {
+    if (unit == 0) unit = 2; // inches
+    if (xres == 0) xres = yres; // if one is zero, set it, too
+    else if (yres == 0) yres = xres;
+    
+    if (unit == 3) { // scale cm to inch
+      xres = xres * 254 / 100;
+      yres = yres * 254 / 100;
+    }
+    
+    // was already set?
+    if (image.resolutionX() == 0 && image.resolutionY() == 0) {
+      image.setResolution(xres, yres);
+    } else {
+      if (image.resolutionX() != xres || image.resolutionY() != yres)
+	std::cerr << "Exif resolution (" << xres << "x" << yres
+		  << ") differs from codec ("
+		  << image.resolutionX() << "x" << image.resolutionY() << ")" << std::endl;
+    }
+  }
   
   exif_rotate(image, orientation);
 }
@@ -512,34 +726,32 @@ void JPEGCodec::parseExif (Image& image)
   struct jpeg_decompress_struct* cinfo = new jpeg_decompress_struct;
   struct my_error_mgr jerr;
   
-  /* Step 1: allocate and initialize JPEG decompression object */
+  // Step 1: allocate and initialize JPEG decompression object
 
-  /* We set up the normal JPEG error routines, then override error_exit. */
+  // We set up the normal JPEG error routines, then override error_exit.
   cinfo->err = jpeg_std_error(&jerr.pub);
   jerr.pub.error_exit = my_error_exit;
-  /* Establish the setjmp return context for my_error_exit to use. */
+  
+  // Establish the setjmp return context for my_error_exit to use.
   if (setjmp(jerr.setjmp_buffer)) {
-    /* If we get here, the JPEG code has signaled an error.
-     * We need to clean up the JPEG object, close the input file, and return.
-     */
+    // If we get here, the JPEG code has signaled an error.
+    // We need to clean up the JPEG object, close the input file, and return.
     jpeg_destroy_decompress (cinfo);
     return;
   }
   
   jpeg_create_decompress (cinfo);
   
-  /* Step 2: specify data source (eg, a file) */
-  
+  // Step 2: specify data source (eg, a file)
   private_copy.seekg (0);
   cpp_stream_src (cinfo, &private_copy);
 
-  /* Step 3: read file parameters with jpeg_read_header() */
-
-  jpeg_read_header(cinfo, TRUE);
+  // Step 3: read file parameters with jpeg_read_header()
+  jpeg_read_header(cinfo, (boolean)TRUE);
   
-  /* Step 4: set parameters for decompression */
+  // Step 4: set parameters for decompression
   
-  cinfo->buffered_image = TRUE; /* select buffered-image mode */
+  cinfo->buffered_image = (boolean)TRUE; // select buffered-image mode
 
   // TODO: set scaling
   if (factor != 1) {
@@ -548,29 +760,31 @@ void JPEGCodec::parseExif (Image& image)
     cinfo->dct_method = JDCT_IFAST;
   }
   
-  /* Step 5: Start decompressor */
+  if (colorspace)
+    cinfo->jpeg_color_space = (J_COLOR_SPACE)colorspace;
+  
+  // Step 5: Start decompressor
   jpeg_start_decompress (cinfo);
   
   image->w = cinfo->output_width;
   image->h = cinfo->output_height;
   
-  /* JSAMPLEs per row in output buffer */
+  // JSAMPLEs per row in output buffer
   int row_stride = cinfo->output_width * cinfo->output_components;
 
   image->resize (image->w, image->h);
   
-  /* Step 6: jpeg_read_scanlines(...); */
+  // Step 6: jpeg_read_scanlines(...)
   
   uint8_t* data = image->getRawData ();
-  JSAMPROW buffer[1];		/* pointer to JSAMPLE row[s] */
+  JSAMPROW buffer[1]; // pointer to JSAMPLE row[s]
   
   while (! jpeg_input_complete(cinfo)) {
     jpeg_start_output(cinfo, cinfo->input_scan_number);
     while (cinfo->output_scanline < cinfo->output_height) {
-      /* jpeg_read_scanlines expects an array of pointers to scanlines.
-       * Here the array is only one element long, but you could ask for
-       * more than one scanline at a time if that's more convenient.
-       */
+      // jpeg_read_scanlines expects an array of pointers to scanlines.
+      // Here the array is only one element long, but you could ask for
+      // more than one scanline at a time if that's more convenient.
       buffer[0] = (JSAMPLE*) data+ (cinfo->output_scanline*row_stride);
       jpeg_read_scanlines(cinfo, buffer, 1);
     }
@@ -599,7 +813,7 @@ bool JPEGCodec::flipY (Image& image)
 
 bool JPEGCodec::rotate (Image& image, double angle)
 {
-  // so roate if the first fraction is zero
+  // so rotate if the first fraction is zero
   switch ((int)(angle * 10)) {
   case 900:  return doTransform (JXFORM_ROT_90, image);
   case 1800: return doTransform (JXFORM_ROT_180, image);
@@ -634,10 +848,10 @@ bool JPEGCodec::toGray (Image& image)
   return doTransform (JXFORM_NONE, image, 0 /* stream */, true /* to gray */);
 }
 
-bool JPEGCodec::scale (Image& image, double xscale, double yscale)
+bool JPEGCodec::scale (Image& image, double xscale, double yscale, bool fixed)
 {
   // we only support fast downscaling
-  if (xscale > 1.0 || yscale > 1.0)
+  if (xscale > 1.0 || yscale > 1.0 || fixed)
     return false; // let the generic scaler handle this
   
   int w_final = (int)(xscale * image.w);
@@ -677,16 +891,15 @@ bool JPEGCodec::readMeta (std::istream* stream, Image& image)
   
   struct my_error_mgr jerr;
   
-  /* Step 1: allocate and initialize JPEG decompression object */
+  // Step 1: allocate and initialize JPEG decompression object
 
-  /* We set up the normal JPEG error routines, then override error_exit. */
+  // We set up the normal JPEG error routines, then override error_exit.
   cinfo->err = jpeg_std_error(&jerr.pub);
   jerr.pub.error_exit = my_error_exit;
-  /* Establish the setjmp return context for my_error_exit to use. */
+  // Establish the setjmp return context for my_error_exit to use.
   if (setjmp(jerr.setjmp_buffer)) {
-    /* If we get here, the JPEG code has signaled an error.
-     * We need to clean up the JPEG object, close the input file, and return.
-     */
+    // If we get here, the JPEG code has signaled an error.
+    // We need to clean up the JPEG object, close the input file, and return.
     jpeg_destroy_decompress (cinfo);
     free (cinfo);
     return false;
@@ -694,20 +907,16 @@ bool JPEGCodec::readMeta (std::istream* stream, Image& image)
   
   jpeg_create_decompress (cinfo);
   
-  /* Step 2: specify data source (eg, a file) */
-  
+  // Step 2: specify data source (eg, a file)
   cpp_stream_src (cinfo, stream);
 
-  /* Step 3: read file parameters with jpeg_read_header() */
-
-  jpeg_read_header(cinfo, TRUE);
+  // Step 3: read file parameters with jpeg_read_header()
+  jpeg_read_header(cinfo, (boolean)TRUE);
   
-  /* Step 4: set parameters for decompression */
-  
-  cinfo->buffered_image = TRUE; /* select buffered-image mode */
-  
-  /* Step 5: Start decompressor */
+  // Step 4: set parameters for decompression
+  cinfo->buffered_image = (boolean)TRUE; /* select buffered-image mode */
   
+  // Step 5: Start decompressor
   jpeg_start_decompress (cinfo);
   
   image.w = cinfo->output_width;
@@ -715,22 +924,22 @@ bool JPEGCodec::readMeta (std::istream* stream, Image& image)
   image.spp = cinfo->output_components;
   image.bps = 8;
   
-  /* These three values are not used by the JPEG code, merely copied */
-  /* into the JFIF APP0 marker. JFIF code for pixel size units. */
+  // These three values are not used by the JPEG code, merely copied
+  // into the JFIF APP0 marker. JFIF code for pixel size units.
   switch (cinfo->density_unit)
     {
-    case 1: /* dots/inch */
+    case 1: // dots/inch
       image.setResolution(cinfo->X_density, cinfo->Y_density);
       break;
-    case 2: /* dots/cm */
+    case 2: // dots/cm
       image.setResolution(cinfo->X_density * 254 / 100,
 			  cinfo->Y_density * 254 / 100);
       break;
-    default: /* 0 for unknown, ratio may still be defined */
+    default: // 0 for unknown, ratio may still be defined
       image.setResolution(0, 0);
     }
   
-  /* This is an important step since it will release a good deal of memory. */
+  // This is an important step since it will release a good deal of memory.
   jpeg_finish_decompress(cinfo);
   jpeg_destroy_decompress(cinfo);
   delete (cinfo);
@@ -743,7 +952,7 @@ bool JPEGCodec::doTransform (JXFORM_CODE code, Image& image,
 			     unsigned int x, unsigned int y,
 			     unsigned int w, unsigned int h)
 {
-  jpeg_transform_info transformoption; /* image transformation options */
+  jpeg_transform_info transformoption = {}; // image transformation options
   
   jpeg_decompress_struct srcinfo;
   jpeg_compress_struct dstinfo;
@@ -751,15 +960,12 @@ bool JPEGCodec::doTransform (JXFORM_CODE code, Image& image,
   
   std::cerr << "Transforming DCT coefficients." << std::endl;
   
-  jvirt_barray_ptr * src_coef_arrays;
-  jvirt_barray_ptr * dst_coef_arrays;
-  
-  /* Initialize the JPEG decompression object with default error handling. */
+  // Initialize the JPEG decompression object with default error handling.
   srcinfo.err = jpeg_std_error(&jsrcerr);
   jpeg_create_decompress(&srcinfo);
-  /* Initialize the JPEG compression object with default error handling. */
+  // Initialize the JPEG compression object with default error handling.
   dstinfo.err = jpeg_std_error(&jdsterr);
-  /* Initialize the JPEG compression object with default error handling. */
+  // Initialize the JPEG compression object with default error handling.
   jpeg_create_compress(&dstinfo);
   
   srcinfo.mem->max_memory_to_use = dstinfo.mem->max_memory_to_use;
@@ -767,15 +973,15 @@ bool JPEGCodec::doTransform (JXFORM_CODE code, Image& image,
   private_copy.seekg (0);
   cpp_stream_src (&srcinfo, &private_copy);
   
-  /* Read file header */
-  jpeg_read_header(&srcinfo, TRUE);
+  // Read file header
+  jpeg_read_header(&srcinfo, (boolean)TRUE);
   
   transformoption.transform = code;
-  transformoption.trim = TRUE;
-  transformoption.perfect = FALSE;
-  transformoption.force_grayscale = to_gray ? TRUE : FALSE;
+  transformoption.trim = (boolean)TRUE;
+  transformoption.perfect = (boolean)FALSE;
+  transformoption.force_grayscale = (boolean)(to_gray ? TRUE : FALSE);
   
-  transformoption.crop = crop ? TRUE : FALSE;
+  transformoption.crop = (boolean)(crop ? TRUE : FALSE);
   if (crop) {
     transformoption.crop_xoffset = x;
     transformoption.crop_xoffset_set = JCROP_POS;
@@ -787,20 +993,19 @@ bool JPEGCodec::doTransform (JXFORM_CODE code, Image& image,
     transformoption.crop_height_set = JCROP_POS;
   }
   
-  /* Any space needed by a transform option must be requested before
-   * jpeg_read_coefficients so that memory allocation will be done right.
-   */
+  // Any space needed by a transform option must be requested before
+  // jpeg_read_coefficients so that memory allocation will be done right.
   jtransform_request_workspace(&srcinfo, &transformoption);
   
-  /* Read source file as DCT coefficients */
-  src_coef_arrays = jpeg_read_coefficients(&srcinfo);
+  // Read source file as DCT coefficients
+  jvirt_barray_ptr* src_coef_arrays = jpeg_read_coefficients(&srcinfo);
   
-  /* Initialize destination compression parameters from source values */
+  // Initialize destination compression parameters from source values
   jpeg_copy_critical_parameters(&srcinfo, &dstinfo);
 
-  /* Adjust destination parameters if required by transform options;
-   * also find out which set of coefficient arrays will hold the output.
-   */
+  // Adjust destination parameters if required by transform options;
+  // also find out which set of coefficient arrays will hold the output.
+  jvirt_barray_ptr* dst_coef_arrays;
   if (transformoption.transform != JXFORM_NONE ||
       transformoption.force_grayscale ||
       transformoption.crop)
@@ -810,7 +1015,7 @@ bool JPEGCodec::doTransform (JXFORM_CODE code, Image& image,
   else
     dst_coef_arrays = src_coef_arrays;
   
-  /* Specify data destination for compression */
+  // Specify data destination for compression
   std::stringstream stream;
   if (!s)
     stream.str().reserve(private_copy.str().size());
@@ -818,15 +1023,15 @@ bool JPEGCodec::doTransform (JXFORM_CODE code, Image& image,
   
   jpeg_compress_set_density (&dstinfo, image);
   
-  /* Start compressor (note no image data is actually written here) */
+  // Start compressor (note no image data is actually written here)
   jpeg_write_coefficients(&dstinfo, dst_coef_arrays);
   
-  /* Execute image transformation, if any */
+  // Execute image transformation, if any
   jtransform_execute_transformation(&srcinfo, &dstinfo,
 				    src_coef_arrays,
 				    &transformoption);
   
-  /* Finish compression and release memory */
+  // Finish compression and release memory
   jpeg_finish_compress(&dstinfo);
   jpeg_destroy_compress(&dstinfo);
   jpeg_finish_decompress(&srcinfo);
@@ -842,50 +1047,33 @@ bool JPEGCodec::doTransform (JXFORM_CODE code, Image& image,
     image.setCodec(this);
     
     // Update meta: w, h, spp might have changed.
-    
+    image.w = transformoption.output_width;
+    image.h = transformoption.output_height;
+
     // avoid the expensive readMeta for some cases
     switch (code) {
     case JXFORM_ROT_90:
     case JXFORM_ROT_270:
-      if (image.w % 8 == 0 && image.h % 8 == 0) {
 	image.setResolution(image.resolutionX(), image.resolutionY());
 	image.setCodec(this);
-	{ int t = image.w; image.w = image.h; image.h = t; }
-	return true;
-      }
-      break;
-      
+
     case JXFORM_ROT_180:
     case JXFORM_FLIP_H:
     case JXFORM_FLIP_V:
-      if (image.w % 8 == 0 && image.h % 8 == 0)
-	return true;
       
     default:
       ; // silence compiler
     }
     
-    if (crop) {
-      if (x % 8 == 0 && y % 8 == 0 &&
-	  w % 8 == 0 && h % 8 == 0)
-	{
-	  image.w = w;
-	  image.h = h;
-	  return true;
-	}
-    }
-    
     if (to_gray) {
       image.spp = 1;
-      return true;
     }
     
     // We re-read the header because we do not want to re-hardcode the
     // trimming required for all the other corner cases.
-    // Tail call as micro optimization to already free tmp objects.
-    std::cerr << "Re-reading meta data." << std::endl;
-    readMeta (&private_copy, image);
-    image.setCodec(this);
+    //std::cerr << "Re-reading meta data." << std::endl;
+    //readMeta (&private_copy, image);
+    //image.setCodec(this);
   }
   
   return true;
