@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2006 - 2010 René Rebe
+ * Copyright (C) 2006 - 2015 René Rebe, ExactCODE GmbH
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,11 +26,14 @@
  */
 
 #include <string.h> // memcpy
-
 #include <iostream>
 #include <string>
 #include <sstream>
 
+#ifdef _MSC_VER
+#include <vector>
+#endif
+
 #include "pnm.hh"
 #include "Endianess.hh"
 using namespace Exact;
@@ -146,7 +149,7 @@ int PNMCodec::readImage (std::istream* stream, Image& image, const std::string&
     }
   else // binary data
     {
-      const int stride = image.stride ();
+      const int stride = image.stride();
       const int bps = image.bps;
       
       for (int y = 0; y < image.h; ++y)
@@ -196,7 +199,6 @@ bool PNMCodec::writeImage (std::ostream* stream, Image& image, int quality,
   if (c == "plain")
     c = "ascii";
 
-  
   if (c != "ascii")
     format += 3;
   
@@ -206,7 +208,8 @@ bool PNMCodec::writeImage (std::ostream* stream, Image& image, int quality,
   *stream << image.w << " " << image.h << std::endl;
   
   // maxval
-  int maxval = (1 << image.bps) - 1;
+  const int maxval = (1 << image.bps) - 1;
+  const int divval = image.bps < 8 ? 255 / maxval : 1;
   
   if (image.bps > 1)
     *stream << maxval << std::endl;
@@ -216,62 +219,60 @@ bool PNMCodec::writeImage (std::ostream* stream, Image& image, int quality,
     {
       for (int y = 0; y < image.h; ++y)
 	{
-	  for (int x = 0; x < image.w; ++x)
+	  it = it.at(0, y);
+	  for (int x = 0; x < image.w; ++x, ++it)
 	    {
-	      *it;
-	      
 	      if (x != 0)
 		*stream << " ";
 	      
+	      *it;
 	      if (image.spp == 1) {
-		int i = it.getL();
+		unsigned i = it.getL();
 
 		// only mode 1 is defined with 1 == black, ...
-                if (format == 1) i = 255 - i;
-
-		*stream << i / (255 / maxval);
+		if (format == 1) i = 255 - i;
+		*stream << i / divval;
 	      }
 	      else {
 		uint16_t r = 0, g = 0, b = 0;
 		it.getRGB (&r, &g, &b);
 		*stream << (int)r << " " << (int)g << " " << (int)b;
 	      }
-	      ++it;
 	    }
 	  *stream << std::endl;
 	}
     }
   else
     {
-      const int stride = image.stride ();
       const int bps = image.bps;
-      
-      uint8_t* ptr = (uint8_t*) malloc (stride);
-      
+      const int stride = image.stride();
+      const int stridefill = image.stridefill();
+#ifndef _MSC_VER
+      uint8_t ptr[stridefill];
+#else
+      std::vector<uint8_t> ptr(stridefill);
+#endif
       for (int y = 0; y < image.h; ++y)
 	{
-	  memcpy (ptr, image.getRawData() + y * stride, stride);
+	  memcpy (&ptr[0], image.getRawData() + y * stride, stridefill);
 	  
 	  // is this publically defined somewhere???
 	  if (bps == 1) {
-	    uint8_t* xor_ptr = ptr;
+	    uint8_t* xor_ptr = &ptr[0];
 	    for (int x = 0; x < image.w; x += 8)
 	      *xor_ptr++ ^= 0xff;
 	  }
 	  
 	  if (bps == 16) {
-	    uint16_t* swap_ptr = (uint16_t*)ptr;
-	    for (int x = 0; x < stride/2; ++x, ++swap_ptr)
+	    uint16_t* swap_ptr = (uint16_t*)&ptr[0];
+	    for (int x = 0; x < stridefill / 2; ++x, ++swap_ptr)
 	      *swap_ptr = ByteSwap<BigEndianTraits, NativeEndianTraits, uint16_t>::Swap (*swap_ptr);
 	  }
 	  
-	  stream->write ((char*)ptr, stride);
+	  stream->write((char*)&ptr[0], stridefill);
 	}
-      free (ptr);
     }
   
-  stream->flush ();
-  
   return true;
 }
 
