@@ -1,6 +1,6 @@
 /*
  * The ExactImage library's displayy compatible command line frontend.
- * Copyright (C) 2006 - 2010 René Rebe
+ * Copyright (C) 2006 - 2016 René Rebe, ExactCODE GmbH
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,6 +31,7 @@
 #include <algorithm>
 #include <iostream>
 #include <sstream>
+#include <unistd.h>
 
 #include "ArgumentList.hh"
 using namespace Utility;
@@ -82,7 +83,10 @@ void Viewer::Zoom (double f)
   
   Evas_Coord w = (Evas_Coord) (zoom * image->w / 100);
   Evas_Coord h = (Evas_Coord) (zoom * image->h / 100);
-  
+  {
+    if (image->resolutionX() > 0 && image->resolutionY() > 0)
+      h = h * image->resolutionX() / image->resolutionY();
+  }  
   evas_bgr_image->Resize (w, h);
   
   // recenter view
@@ -134,8 +138,8 @@ void Viewer::UpdateOSD (const std::string& str1, const std::string& str2)
 {
   evas_osd_text1->Text (str1);
   evas_osd_text2->Text (str2);
-  evas_osd_rect->Resize (std::max(evas_osd_text1->Width(), evas_osd_text2->Width()) + 4,
-			 evas_osd_text1->Height() + evas_osd_text2->Height() + 8);
+  evas_osd_rect->Resize (std::max(evas_osd_text1->Width(), evas_osd_text2->Width()) + 8*HiDPI,
+			 evas_osd_text1->Height() + evas_osd_text2->Height() + 8*HiDPI);
   AlphaOSD (0xFF);
   osd_timer.Reset ();
 }
@@ -185,6 +189,13 @@ int Viewer::Run (bool opengl)
 
   Window root = RootWindow(dpy, scr);
 
+  int xres = 25.4 * XDisplayWidth(dpy, scr) / XDisplayWidthMM(dpy, scr) + .5;
+  int yres = 25.4 * XDisplayHeight(dpy, scr) / XDisplayHeightMM(dpy, scr) + .5;
+  if (xres >= 192 || yres >= 192)
+    HiDPI = 2;
+  else if (xres >= 144 || yres >= 144)
+    HiDPI = 1.5;
+  
   attr.backing_store = NotUseful;
   attr.border_pixel = 0;
   attr.background_pixmap = None;
@@ -317,14 +328,14 @@ int Viewer::Run (bool opengl)
   
   evas_osd_text1 = new EvasText (*evas);
   evas_osd_text1->Layer (10);
-  evas_osd_text1->Move (4,4);
-  evas_osd_text1->Font ("Vera", 10);
+  evas_osd_text1->Move (4*HiDPI, 4*HiDPI);
+  evas_osd_text1->Font ("Vera", 10*HiDPI);
   evas_osd_text1->Show ();
 
   evas_osd_text2 = new EvasText (*evas);
   evas_osd_text2->Layer (10);
-  evas_osd_text2->Move (4,8 + evas_osd_text1->Height());
-  evas_osd_text2->Font ("Vera", 10);
+  evas_osd_text2->Move (4*HiDPI, 8*HiDPI + evas_osd_text1->Height());
+  evas_osd_text2->Font ("Vera", 10*HiDPI);
   evas_osd_text2->Show ();
 
   evas_osd_rect = new EvasRectangle (*evas);
@@ -689,7 +700,7 @@ void Viewer::ImageToEvas ()
     evas_image->SmoothScale (false);
     evas_image->Layer (1);
     evas_image->Move (0, 0);
-    evas_image->Resize (image->w,image->h);
+    evas_image->Resize (image->w, image->h);
     evas_content.push_back(evas_image);
   } else {
     evas_image->SetData (0);
@@ -706,72 +717,69 @@ void Viewer::ImageToEvas ()
 
 EvasImage* Viewer::ImageToEvas (Image* image, EvasImage* eimage)
 {
-  uint8_t* evas_data = 0;
-  
   if (!eimage) {
     eimage = new EvasImage (*evas);
     eimage->SmoothScale (false);
     eimage->Layer (1);
     eimage->Move (0, 0);
-    eimage->Resize (image->w,image->h);
+    eimage->Resize (image->w, image->h);
   } else {
     eimage->SetData (0);
   }
   
-  evas_data = (uint8_t*) realloc (evas_data, image->w*image->h*4);
-  uint8_t* src_ptr = image->getRawData();
-  uint8_t* dest_ptr = evas_data;
-
+  evas_data = (uint8_t*)realloc(evas_data, image->w * image->h * 4);
+  uint8_t* src_data = image->getRawData();
+  const unsigned src_stride = image->stride();
+  const unsigned dst_stride = image->w * 4;
   const int spp = image->spp; 
   
-  if (channel == 0)
-    {
-      for (int y=0; y < image->h; ++y)
-        for (int x=0; x < image->w; ++x, dest_ptr +=4, src_ptr += spp) {
-          if (!Exact::NativeEndianTraits::IsBigendian) {
-            dest_ptr[0] = src_ptr[2];
-            dest_ptr[1] = src_ptr[1];
-            dest_ptr[2] = src_ptr[0];
-            if (spp == 4)
-              dest_ptr[3] = src_ptr[3]; // alpha
-          }
-          else {
-            dest_ptr[1] = src_ptr[0];
-            dest_ptr[2] = src_ptr[1];
-            dest_ptr[3] = src_ptr[2];
-            if (spp == 4)
-              dest_ptr[0] = src_ptr[3]; // alpha
-          }
-        }
+  for (int y = 0; y < image->h; ++y) {
+    uint8_t* src_ptr = src_data + y * src_stride;
+    uint8_t* dest_ptr = evas_data + y * dst_stride;
+    
+    if (channel == 0) {
+      for (int x = 0; x < image->w; ++x, dest_ptr +=4, src_ptr += spp) {
+	if (!Exact::NativeEndianTraits::IsBigendian) {
+	  dest_ptr[0] = src_ptr[2];
+	  dest_ptr[1] = src_ptr[1];
+	  dest_ptr[2] = src_ptr[0];
+	  if (spp == 4)
+	    dest_ptr[3] = src_ptr[3]; // alpha
+	}
+	else {
+	  dest_ptr[1] = src_ptr[0];
+	  dest_ptr[2] = src_ptr[1];
+	  dest_ptr[3] = src_ptr[2];
+	  if (spp == 4)
+	    dest_ptr[0] = src_ptr[3]; // alpha
+	}
+      }
     }
-  else
-    {
+    else { // channel
       bool intensity = channel > 3;
       int ch = (channel-1) % 3;
       
-      for (int y=0; y < image->h; ++y)
-        for (int x=0; x < image->w; ++x, dest_ptr +=4, src_ptr += spp) {
-          if (!Exact::NativeEndianTraits::IsBigendian) {
-            dest_ptr[0] = dest_ptr[1] = dest_ptr[2] =
-              intensity ? src_ptr[ch] : 0;
-            if (!intensity)
-	      dest_ptr[2-ch] = src_ptr[ch];
-          }
-          else {
-            dest_ptr[1] = dest_ptr[2] = dest_ptr[3] =
-              intensity ? src_ptr[ch] : 0;
-            if (!intensity)
-              dest_ptr[1+ch] = src_ptr[ch];
-          }
-        }
+      for (int x=0; x < image->w; ++x, dest_ptr +=4, src_ptr += spp) {
+	if (!Exact::NativeEndianTraits::IsBigendian) {
+	  dest_ptr[0] = dest_ptr[1] = dest_ptr[2] = intensity ? src_ptr[ch] : 0;
+	  if (!intensity)
+	    dest_ptr[2-ch] = src_ptr[ch];
+	}
+	else {
+	  dest_ptr[1] = dest_ptr[2] = dest_ptr[3] = intensity ? src_ptr[ch] : 0;
+	  if (!intensity)
+	    dest_ptr[1+ch] = src_ptr[ch];
+	}
+      }
     }
+  }
   
   eimage->Alpha (spp == 4);
   eimage->Resize (image->w, image->h);
   eimage->ImageSize (image->w, image->h);
-  eimage->ImageFill (0, 0, image->w,image->h);
+  eimage->ImageFill (0, 0, image->w, image->h);
   eimage->SetData (evas_data);
-  eimage->DataUpdateAdd (0, 0, image->w,image->h);
+  eimage->DataUpdateAdd (0, 0, image->w, image->h);
   eimage->Show ();
   
   return eimage;
@@ -787,7 +795,7 @@ int main (int argc, char** argv)
   ArgumentList arglist (true); // enable residual gathering
   
   // setup the argument list
-  Argument<bool> arg_help ("", "help",
+  Argument<bool> arg_help ("h", "help",
                            "display this help text and exit");
   arglist.Add (&arg_help);
 #if EVASGL == 1
@@ -799,9 +807,8 @@ int main (int argc, char** argv)
   // parse the specified argument list - and maybe output the Usage
   if (!arglist.Read (argc, argv) || arg_help.Get() == true || arglist.Residuals().empty())
     {
-      cerr << "Exact image viewer (edisplay)."
-	   << endl << "Version " VERSION
-	   <<  " - Copyright (C) 2006 - 2010 by René Rebe" << std::endl
+      cerr << "ExactImage viewer - edisplay, version " VERSION << std::endl
+	   << "Copyright (C) 2006 - 2016 René Rebe, ExactCODE GmbH" << std::endl
 	   << "Usage:" << endl;
       
       arglist.Usage (cerr);
