@@ -1,6 +1,6 @@
 /*
  * The ExactImage library's convert compatible command line frontend.
- * Copyright (C) 2005 - 2011 René Rebe, ExactCODE GmbH
+ * Copyright (C) 2005 - 2015 René Rebe, ExactCODE GmbH
  * Copyright (C) 2005, 2008 Archivista GmbH
  * 
  * This program is free software; you can redistribute it and/or modify
@@ -87,7 +87,7 @@ static void freeImages()
   }
 }
 
-Argument<int> arg_quality ("", "quality",
+Argument<int> arg_quality ("q", "quality",
 			   "quality setting used for writing compressed images\n\t\t"
 			   "integer range 0-100, the default is 75",
 			   0, 1, true, true);
@@ -185,12 +185,14 @@ bool convert_output (const Argument<std::string>& arg)
   std::string compression = "";
   if (arg_compression.Size())
     compression = arg_compression.Get();
+  Args args(compression);
+  const bool append = args.contains("append");
   
   int i = 0, f = 0;
   images_iterator it = images.begin();
   ImageCodec* codec = 0;
   std::fstream* stream = 0;
-
+  
   for (; it != images.end(); ++it)
     {
       if (!codec)
@@ -203,10 +205,12 @@ bool convert_output (const Argument<std::string>& arg)
 	  std::string file = arg.Get(f);
 	  std::string cod = ImageCodec::getCodec(file);
 	  std::string ext = ImageCodec::getExtension(file);
-	  stream = new std::fstream(file.c_str(),
-				    std::ios::in | std::ios::out | std::ios::trunc);
+	  stream =
+	    new std::fstream(file.c_str(),
+			     std::ios::in | std::ios::out | std::ios::binary |
+			     (append ? (std::ios_base::openmode)0 : std::ios::trunc));
 	  
-	  codec = ImageCodec::MultiWrite(stream, cod, ext);
+	  codec = ImageCodec::MultiWrite(stream, cod, ext, compression);
 	  // if we got no codec, write a classic, single-page file
 	  if (!codec) {
 	    if (!ImageCodec::Write(stream, **it, cod, ext, quality, compression))
@@ -326,17 +330,39 @@ bool convert_blur (const Argument<double>& arg)
   return true;
 }
 
-bool convert_scale (const Argument<double>& arg)
+bool parse_scale(const Argument<std::string>& arg, double& sx, double& sy, bool& fixed)
 {
-  double f = arg.Get();
-  FOR_ALL_IMAGES(scale, f, f);
+  // TODO: pretty C++ parser
+  int n;
+  if ((n = sscanf(arg.Get().c_str(), "%lfx%lf", &sx, &sy))) {
+    fixed = true;
+    if (n < 2) {
+       sy = sx;
+       fixed = false;
+    }
+    return true;
+  }
+  return false;
+}
+
+bool convert_scale (const Argument<std::string>& arg)
+{
+  bool fixed; double sx, sy;
+  if (!parse_scale(arg, sx, sy, fixed)) {
+    return false;
+  }
+  FOR_ALL_IMAGES(scale,  sx, sy, fixed);
   return true;
 }
 
-bool convert_thumbnail_scale (const Argument<double>& arg)
+bool convert_thumbnail_scale (const Argument<std::string>& arg)
 {
-  double f = arg.Get();
-  FOR_ALL_IMAGES(thumbnail_scale, f, f);
+  bool fixed; double sx, sy;
+  if (!parse_scale(arg, sx, sy, fixed)) {
+    std::cerr << "scale '" << arg.Get() << "' could not be parsed." << std::endl;
+    return false;
+  }
+  FOR_ALL_IMAGES(thumbnail_scale,  sx, sy, fixed);
   return true;
 }
 
@@ -361,38 +387,58 @@ bool convert_lightness (const Argument<double>& arg)
   return true;
 }
 
-bool convert_nearest_scale (const Argument<double>& arg)
+bool convert_nearest_scale (const Argument<std::string>& arg)
 {
-  double f = arg.Get();
-  FOR_ALL_IMAGES(nearest_scale, f, f);
+  bool fixed; double sx, sy;
+  if (!parse_scale(arg, sx, sy, fixed)) {
+    std::cerr << "scale '" << arg.Get() << "' could not be parsed." << std::endl;
+    return false;
+  }
+  FOR_ALL_IMAGES(nearest_scale, sx, sy, fixed);
   return true;
 }
 
-bool convert_bilinear_scale (const Argument<double>& arg)
+bool convert_bilinear_scale (const Argument<std::string>& arg)
 {
-  double f = arg.Get();
-  FOR_ALL_IMAGES(bilinear_scale, f, f);
+  bool fixed; double sx, sy;
+  if (!parse_scale(arg, sx, sy, fixed)) {
+    std::cerr << "scale '" << arg.Get() << "' could not be parsed." << std::endl;
+    return false;
+  }
+  FOR_ALL_IMAGES(bilinear_scale, sx, sy, fixed);
   return true;
 }
 
-bool convert_bicubic_scale (const Argument<double>& arg)
+bool convert_bicubic_scale (const Argument<std::string>& arg)
 {
-  double f = arg.Get();
-  FOR_ALL_IMAGES(bicubic_scale, f, f);
+  bool fixed; double sx, sy;
+  if (!parse_scale(arg, sx, sy, fixed)) {
+    std::cerr << "scale '" << arg.Get() << "' could not be parsed." << std::endl;
+    return false;
+  }
+  FOR_ALL_IMAGES(bicubic_scale, sx, sy, fixed);
   return true;
 }
 
-bool convert_box_scale (const Argument<double>& arg)
+bool convert_box_scale (const Argument<std::string>& arg)
 {
-  double f = arg.Get();
-  FOR_ALL_IMAGES(box_scale, f, f);
+  bool fixed; double sx, sy;
+  if (!parse_scale(arg, sx, sy, fixed)) {
+    std::cerr << "scale '" << arg.Get() << "' could not be parsed." << std::endl;
+    return false;
+  }
+  FOR_ALL_IMAGES(box_scale, sx, sy, fixed);
   return true;
 }
 
-bool convert_ddt_scale (const Argument<double>& arg)
+bool convert_ddt_scale (const Argument<std::string>& arg)
 {
-  double f = arg.Get();
-  FOR_ALL_IMAGES(ddt_scale, f, f);
+  bool fixed; double sx, sy;
+  if (!parse_scale(arg, sx, sy, fixed)) {
+    std::cerr << "scale '" << arg.Get() << "' could not be parsed." << std::endl;
+    return false;
+  }
+  FOR_ALL_IMAGES(ddt_scale, sx, sy, fixed);
   return true;
 }
 
@@ -432,32 +478,14 @@ bool convert_convolve (const Argument<double>& arg)
 
 bool convert_dither_floyd_steinberg (const Argument<int>& arg)
 {
-  bool ret = true;
-  for (images_iterator it = images.begin(); it != images.end(); ++it)
-    {
-      if ((*it)->bps != 8) {
-	std::cerr << "Can only dither 8 bit data right now." << std::endl;
-	ret = false;
-      }
-      else
-	FloydSteinberg((*it)->getRawData(), (*it)->w, (*it)->h, arg.Get(), (*it)->spp);
-    }
-  return ret;
+  FOR_ALL_IMAGES(FloydSteinberg, arg.Get());
+  return true; 
 }
 
 bool convert_dither_riemersma (const Argument<int>& arg)
 {
-  bool ret = true;
-  for (images_iterator it = images.begin(); it != images.end(); ++it)
-    {
-      if ((*it)->bps != 8) {
-	std::cerr << "Can only dither 8 bit data right now." << std::endl;
-	ret = false;
-      }
-      else
-	Riemersma ((*it)->getRawData(), (*it)->w, (*it)->h, arg.Get(), (*it)->spp);
-    }
-  return ret;
+  FOR_ALL_IMAGES(Riemersma, arg.Get());
+  return true;
 }
 
 bool convert_edge (const Argument<bool>& arg)
@@ -470,6 +498,18 @@ bool convert_edge (const Argument<bool>& arg)
   return true;
 }
 
+bool convert_pop (const Argument<bool>& arg)
+{
+  if (!images.empty()) {
+    delete(images.back());
+    images.pop_back();
+    return true;
+  } else {
+    std::cerr << "no image on stack to pop" << std::endl;
+    return false;
+  }
+}
+ 
 bool convert_resolution (const Argument<std::string>& arg)
 {
   int xres, yres, n;
@@ -498,15 +538,15 @@ bool convert_size (const Argument<std::string>& arg)
     {
       // this is mostly used to set the size for raw data loads
       // so we need to have at least one (empty) image
-      if (images.empty())
-	images.push_back(new Image);
+      if (images.empty()) {
+	Image* image = new Image;
+	image->spp = image->bps = 1;
+	images.push_back(image);
+      }
       
-      for (images_iterator it = images.begin(); it != images.end(); ++it)
-	{
-	  (*it)->w = w;
-	  (*it)->h = h;
-	  (*it)->setRawData(0);
-	}
+      for (images_iterator it = images.begin(); it != images.end(); ++it) {
+	(*it)->resize(w, h);
+      }
       
       return true;
     }
@@ -668,9 +708,9 @@ bool convert_foreground (const Argument<std::string>& arg)
 
 bool convert_line (const Argument<std::string>& arg)
 {
-  unsigned int x1, y1, x2, y2;
+  float x1, y1, x2, y2;
   
-  if (sscanf(arg.Get().c_str(), "%d,%d,%d,%d", &x1, &y1, &x2, &y2) == 4)
+  if (sscanf(arg.Get().c_str(), "%f,%f,%f,%f", &x1, &y1, &x2, &y2) == 4)
     {
       Path path;
       path.moveTo (x1, y1);
@@ -914,45 +954,45 @@ int main (int argc, char* argv[])
   arglist.Add (&arg_blur);
 
   
-  Argument<double> arg_scale ("", "scale",
+  Argument<std::string> arg_scale ("", "scale",
 			      "scale image data using a method suitable for specified factor",
-			      0.0, 0, 1, true, true);
+			      0, 1, true, true);
   arg_scale.Bind (convert_scale);
   arglist.Add (&arg_scale);
   
-  Argument<double> arg_nearest_scale ("", "nearest-scale",
+  Argument<std::string> arg_nearest_scale ("", "nearest-scale",
 				   "scale image data to nearest neighbour",
-				   0.0, 0, 1, true, true);
+				   0, 1, true, true);
   arg_nearest_scale.Bind (convert_nearest_scale);
   arglist.Add (&arg_nearest_scale);
 
-  Argument<double> arg_bilinear_scale ("", "bilinear-scale",
+  Argument<std::string> arg_bilinear_scale ("", "bilinear-scale",
 				       "scale image data with bi-linear filter",
-				       0.0, 0, 1, true, true);
+				       0, 1, true, true);
   arg_bilinear_scale.Bind (convert_bilinear_scale);
   arglist.Add (&arg_bilinear_scale);
 
-  Argument<double> arg_bicubic_scale ("", "bicubic-scale",
+  Argument<std::string> arg_bicubic_scale ("", "bicubic-scale",
 				      "scale image data with bi-cubic filter",
-				      0.0, 0, 1, true, true);
+				      0, 1, true, true);
   arg_bicubic_scale.Bind (convert_bicubic_scale);
   arglist.Add (&arg_bicubic_scale);
 
-  Argument<double> arg_ddt_scale ("", "ddt-scale",
+  Argument<std::string> arg_ddt_scale ("", "ddt-scale",
 				      "scale image data with data dependent triangulation",
-				      0.0, 0, 1, true, true);
+				     0, 1, true, true);
   arg_ddt_scale.Bind (convert_ddt_scale);
   arglist.Add (&arg_ddt_scale);
   
-  Argument<double> arg_box_scale ("", "box-scale",
+  Argument<std::string> arg_box_scale ("", "box-scale",
 				   "(down)scale image data with box filter",
-				  0.0, 0, 1, true, true);
+				  0, 1, true, true);
   arg_box_scale.Bind (convert_box_scale);
   arglist.Add (&arg_box_scale);
 
-  Argument<double> arg_thumbnail_scale ("", "thumbnail",
+  Argument<std::string> arg_thumbnail_scale ("", "thumbnail",
 					"quick and dirty down-scale for a thumbnail",
-					0.0, 0, 1, true, true);
+					0, 1, true, true);
   arg_thumbnail_scale.Bind (convert_thumbnail_scale);
   arglist.Add (&arg_thumbnail_scale);
 
@@ -998,6 +1038,14 @@ int main (int argc, char* argv[])
 			   0, 0, true, true);
   arg_edge.Bind (convert_edge);
   arglist.Add (&arg_edge);
+
+
+  Argument<bool> arg_pop ("", "pop",
+                           "pop image from stack",
+                           0, 0, true, true);
+  arg_pop.Bind (convert_pop);
+  arglist.Add (&arg_pop);
+
   
   Argument<std::string> arg_resolution ("", "resolution",
 					"set meta data resolution in dpi to x[xy] e.g. 200 or 200x400",
@@ -1075,8 +1123,8 @@ int main (int argc, char* argv[])
   if (argc == 1 || arg_help.Get() == true)
     {
       std::cerr << "ExactImage converter, version " VERSION << std::endl
-		<< "Copyright (C) 2005 - 2011 René Rebe, ExactCODE" << std::endl
-		<< "Copyright (C) 2005, 2008 Archivista" << std::endl
+		<< "Copyright (C) 2005 - 2015 René Rebe, ExactCODE GmbH" << std::endl
+		<< "Copyright (C) 2005, 2008 Archivista GmbH" << std::endl
 		<< "Usage:" << std::endl;
       
       arglist.Usage (std::cerr);
